module Context where
import Syntax 

--                     b    em   i    s    tt   u   size color
 
data Context = Context !Bool !Bool !Bool !Bool !Bool !Int !Int !Color deriving (Show, Eq)

data Color = Color !Char deriving (Show, Eq)

enterTag :: Context -> Tag -> Context
-- When entering a B tag, the b attribute must be set to true
enterTag (Context b em i s tt u size color) B = Context True em i s tt u size color

-- When entering an EM tag, we must do nothing if the S flag is set
-- otherwise, the em flag must be inverted
enterTag con@(Context b em i True tt u size color) EM = con
enterTag (Context b em i False tt u size color) EM = Context b (not em) i False tt u size color

-- When entering an I tag, the i attribute must be set
enterTag (Context b em i s tt u size color) I = Context b em True s tt u size color

-- When entering a PL tag, the u,s,em,i,b and tt flags must all be set to False
enterTag (Context b em i s tt u size color) PL = Context False False False False False 0 size color

-- When entering a S tag, the s flag must be set, and the em flag must be unset
enterTag (Context b em i s tt u size color) S = Context b False i True tt u size color

-- when entering a TT tag, the tt attribute must be set to true
enterTag (Context b em i s tt u size color) TT = Context b em i s True u size color

-- when entering U tag, we must increase the u attribute
-- we shall increase it only to 3, and 3 shall be that maximum number assigned to u
-- 4 shall thou not assign to u, nor shall thou set it to -1
enterTag (Context b em i s tt 3 size color) U = Context b em i s tt 3 size color
enterTag (Context b em i s tt u size color) U = Context b em i s tt (u+1) size color

-- When entering an SZ tag, we change the size
enterTag (Context b em i s tt u size color) (Sz x) = Context b em i s tt u x color

-- When entering a C tag, we set the color attribute accordingly
enterTag (Context b em i s tt u size color) (C x) = Context b em i s tt u size (Color x)





-- Check whether a given tag would change the given context
changes :: Tag -> Context -> Bool
changes tag context = not (context==(enterTag context tag))

-- let's see if the changes function does what is has to do

testChanges :: Bool

testChanges = if (changes B (Context True True True True True 0 0 (Color 'e'))==True) then False
			  else (if (changes B (Context False True True True True 0 0 (Color 'e'))==False) then False
			        else True
                   )



getB :: Context -> Bool
getB (Context b em i s tt u size color) = b

getEM :: Context -> Bool
getEM (Context b em i s tt u size color) = em

getI :: Context -> Bool
getI (Context b em i s tt u size color) = i

getS :: Context -> Bool
getS (Context b em i s tt u size color) = s

getTT :: Context -> Bool
getTT (Context b em i s tt u size color) = tt

getU :: Context -> Int
getU (Context b em i s tt u size color) = u

getSIZE :: Context -> Int
getSIZE (Context b em i s tt u size color) = size

getCOLOR :: Context -> Color
getCOLOR (Context b em i s tt u size color) = color

emptyContext :: Context
emptyContext = (Context False False False False False 0 (-1) (Color '*'))



-- try to reach a given context from the toplevel
-- there are 3 independent parts:
--   1) b em i s tt u
--   2) size
--   3) color
--
-- that means we can treat them seperatly and append the results
genTags :: Context -> [Tag]
genTags ct = (genSizeTags ct) ++ (genColorTags ct) ++ (genOtherTags ct)

-- generate the tag to set the size. only 1 tag required
genSizeTags :: Context -> [Tag]
genSizeTags (Context b em i s tt u (-1) color) = []
genSizeTags (Context b em i s tt u size color) = [Sz size]


-- generate the tag to set the color. only 1 tag required
genColorTags :: Context -> [Tag]
genColorTags (Context b em i s tt u size (Color '*')) = []
genColorTags (Context b em i s tt u size (Color x)) = [C x]

-- generate the tags to set b,em.i,s,tt and u
genOtherTags :: Context -> [Tag]
-- if b must be true, return <B> + tags that are needed when B should not have been set to true
genOtherTags (Context True em i s tt u size color) = B:genOtherTags(Context False em i s tt u size color)
-- B is False here, same for I and TT
genOtherTags (Context False em True s tt u size color) = I:genOtherTags(Context False em False s tt u size color)
-- B and I are false here
genOtherTags (Context False em False s True u size color)= TT:genOtherTags(Context False em False s False u size color)
-- B, I and TT are false here
-- now the U tags
genOtherTags (Context False em False s False 3 size color)= U:U:U:genOtherTags(Context False em False s False 0 size color)
genOtherTags (Context False em False s False 2 size color)= U:U:genOtherTags(Context False em False s False 0 size color)
genOtherTags (Context False em False s False 1 size color)= U:genOtherTags(Context False em False s False 0 size color)
-- B, I and TT are false, U=0
-- now there are 3 possibilities
-- S = True and EM = False -> <S>
-- S = False and EM = True -> <EM>
-- S = False and EM = False -> no tags 
genOtherTags (Context False False False True False 0 size color)= [S]
genOtherTags (Context False True False False False 0 size color)= [EM]
genOtherTags (Context False False False False False 0 size color)= []
-- if we reach the next clause, something has gone wrong
genOtherTags con = error "no pattern match in genOtherTags"
