module DeltaContext where

import Syntax
import Context

data DeltaContext = DeltaContext !Bool 		-- PL
				 !Bool 		-- S
				 !Bool 		-- B
				 !Bool 		-- I
				 !Bool		-- EM
				 !Bool 		-- TT
				 !Int  		-- U
				 !(Maybe Char) 	-- color
				 !(Maybe Int)	-- size
				 deriving (Show,Eq)

wssclone :: DeltaContext -> DeltaContext
wssclone (DeltaContext pl s b i em tt u c sz) = DeltaContext pl s b i em False 0 c sz

emptyDeltaContext :: DeltaContext
emptyDeltaContext = DeltaContext False False False False False False 0 Nothing Nothing

enterDCTag :: Tag -> DeltaContext -> DeltaContext

enterDCTag (Sz _) dc@(DeltaContext _ _ _ _ _ _ _ _ (Just _)) = dc
enterDCTag (Sz sz) (DeltaContext pl s b i em tt u c Nothing) = DeltaContext pl s b i em tt u c (Just sz)

enterDCTag (C _) dc@(DeltaContext _ _ _ _ _ _ _ (Just _) _) = dc
enterDCTag (C c) (DeltaContext pl s b i em tt u Nothing sz) = DeltaContext pl s b i em tt u (Just c) sz

enterDCTag _ dc@(DeltaContext True _ _ _ _ _ _ _ _) = dc

enterDCTag PL (DeltaContext _ s b i em tt u c sz) = DeltaContext True s b i em tt u c sz

enterDCTag S dc@(DeltaContext _ True _ _ _ _ _ _ _) = dc
enterDCTag S (DeltaContext pl _ b i _ tt u c sz) = DeltaContext pl True b i False tt u c sz

enterDCTag B dc@(DeltaContext _ _ True _ _ _ _ _ _) = dc
enterDCTag B (DeltaContext pl s _ i em tt u c sz) = DeltaContext pl s True i em tt u c sz

enterDCTag I dc@(DeltaContext _ _ _ True _ _ _ _ _) = dc
enterDCTag I (DeltaContext pl s b _ em tt u c sz) = DeltaContext pl s b True em tt u c sz

enterDCTag TT dc@(DeltaContext _ _ _ _  _ True _ _ _) = dc
enterDCTag TT (DeltaContext pl s b i em _ u c sz) = DeltaContext pl s b i em True u c sz

enterDCTag U dc@(DeltaContext _ _ _ _  _ _ 3 _ _) = dc
enterDCTag U (DeltaContext pl s b i em tt u c sz) = DeltaContext pl s b i em tt (u+1) c sz

enterDCTag EM dc@(DeltaContext _ True _ _  _ _ _ _ _) = dc
enterDCTag EM (DeltaContext pl s b i em tt u c sz) = DeltaContext pl s b i (not em) tt u c sz

buildDCEl :: DeltaContext -> Context -> Document -> Document
buildDCEl dc c doc = let tags = mkDCTags dc c
		     in applyTags tags doc

mkDCTags :: DeltaContext -> Context -> [Tag]
mkDCTags (DeltaContext pl s b i em tt u c sz) (Context b' em' i' s' tt' u' sz' (Color c')) =
  mkPL pl b' i' tt' em' s' u' $ mkS s (ifnot pl s') $ mkB b (ifnot pl b') $ mkI i (ifnot pl i') $ mkEM em (ifnot pl (ifnot s' em')) $ mkTT tt (ifnot pl tt') $ mkU u (if pl then 0 else u') $ mkC c c'$ mkSz  sz sz' $ []
  where mkB :: Bool -> Bool -> [Tag] -> [Tag]
        mkB False _ 	= id
	mkB True True 	= id
	mkB _    _      = (B:)

	mkI :: Bool -> Bool -> [Tag] -> [Tag]
	mkI False _	= id
	mkI True  True	= id
	mkI _     _	= (I:)

	mkS :: Bool -> Bool -> [Tag] -> [Tag]
	mkS False _	= id
	mkS True  True	= id
	mkS _     _	= (S:)

	mkTT :: Bool -> Bool -> [Tag] -> [Tag]
	mkTT False _	= id
	mkTT True  True	= id 
	mkTT _     _	= (TT:)

	mkC :: (Maybe Char) -> Char -> [Tag] -> [Tag]
	mkC Nothing _	= id
	mkC (Just c) c' = if c == c' then id else ((C c):)
	
	mkSz :: (Maybe Int) -> Int -> [Tag] -> [Tag]
	mkSz Nothing  _	= id
	mkSz (Just i) i'= if i == i' then id else ((Sz i):)

	mkU :: Int -> Int -> [Tag] -> [Tag]
	mkU 0 _  = id
	mkU u u' = f (min (3 - u') u)
	           where f 3 = (U:) . (U:) . (U:)
		         f 2 = (U:) . (U:)
			 f 1 = (U:)
		   	 f _ = id
	
	mkEM :: Bool -> Bool -> [Tag] -> [Tag]
	mkEM False _	= id
	mkEM True  True = id
	mkEM _     _    = (EM:)

	mkPL :: Bool -> Bool -> Bool -> Bool -> Bool -> Bool -> Int -> [Tag] -> [Tag] 
	mkPL False _     _     _     _     _     _ = id
	mkPL True  False False False False False 0 = id
	mkPL _     _     _     _     _     _     _ = (PL:)

applyTags :: [Tag] -> Document -> Document
applyTags [] doc 	= doc
applyTags (t:ts) doc	= [Tag t (applyTags ts doc)]

xor :: Bool -> Bool -> Bool
xor x y = if x then not y else y

mergeByDeltaContext ::
	(DeltaContext,Document) -> -- left
	(DeltaContext,Document) -> -- right
	Context			-> -- common parent context
	Either ((DeltaContext,Document),(DeltaContext,Document)) (DeltaContext,Document)    -- up

mergeByDeltaContext x@(DeltaContext pl s b i em tt u c sz, doc1) y@(DeltaContext pl' s' b' i' em' tt' u' c' sz', doc2) ct =
 	let plok	= pl == pl'
	    --eq		= plok && (s==s') && (b==b') && (i==i') && (em==em') && (tt=tt') && (u==u')
	    pl'' 	= pl && pl'
	    s''		= s  && s'  && plok
	    b''		= b  && b'  && plok
	    i''		= i  && i'  && plok
	    em''	= em && em' && plok
	    tt''	= tt && tt' && plok
	    u''		= if plok then u `min` u' else 0
	    c''		= c `cand` c'
	    sz''	= sz `cand` sz
	    f x y       = if x == y then Nothing else x
	    ct'		= updateContext ct pl'' s'' b'' i'' em'' tt'' u'' c'' sz''
	    lct		= DeltaContext (pl  `xor` pl'') (s  `xor` s'') (b  `xor` b'') (i  `xor` i'') (em  `xor` em'') (tt  `xor` tt'') (u  - u'') (f c  c'') (f sz  sz'')
	    l		= buildDCEl lct ct' doc1
	    r		= buildDCEl (DeltaContext (pl' `xor` pl'') (s' `xor` s'') (b' `xor` b'') (i' `xor` i'') (em' `xor` em'') (tt' `xor` tt'') (u' - u'') (f c' c'') (f sz' sz'')) ct' doc2
	    ndc		= DeltaContext pl'' s'' b'' i'' em'' tt'' u'' c'' sz'' 
	in if ndc == emptyDeltaContext then Left (x, y)
	                               else Right (ndc,  r++l)
	    

cand :: Maybe a -> Maybe a -> Maybe a
cand Nothing    _	= Nothing
cand _          Nothing	= Nothing
cand c@(Just _) _	= c

ifnot :: Bool -> Bool -> Bool
ifnot True _ = False
ifnot _    x = x

updateContext :: Context -> Bool -> Bool -> Bool -> Bool -> Bool -> Bool -> Int -> Maybe Char -> Maybe Int -> Context
updateContext ct pl s b i em tt u c sz =
	ucSz sz $ ucC c $ ucU u $ ucTT tt $ ucEM em $ ucI i $ ucB b $ ucS s $ ucPL pl ct
	where ucPL x ct = if x then enterTag ct PL else ct
	      ucS  x ct = if x then enterTag ct S  else ct
	      ucB  x ct = if x then enterTag ct B  else ct
	      ucI  x ct = if x then enterTag ct I  else ct
	      ucEM x ct' = if x then enterTag ct' EM else ct'
	      ucTT x ct = if x then enterTag ct TT else ct
	      ucU  0 ct = ct
	      ucU  x ct = ucU (x-1) (enterTag ct U)
	      ucC Nothing ct  = ct
	      ucC (Just c) ct = enterTag ct (C c)
	      ucSz Nothing ct = ct
	      ucSz (Just sz) ct = enterTag ct (Sz sz)
