module Parser where

import Syntax

import Context

import DeltaContext

import IO

import Monad

-- only for hugs
--isEOF = hugsIsEOF

type Env = Maybe (DeltaContext, Document)

parse :: IO Document

parse = do
	   [doc] <- ps [[]] [] [emptyContext] [Nothing] [True]
	   return doc


ps :: 
	[Document]	-> -- stack of parent dccuments
	--[Document] 			-> -- stack of parent documents
	[Tag] 				-> -- stack of parent tags
	[Context] 			-> -- stack of contexts
	[Env] 				-> -- stack of extra information
	[Bool]				-> -- all white spaces
	IO [Document]
	
ps docs tags ct env w = do 
                     	   eof <- isEOF
		     	   if eof then return (removeEnv env ct docs)
		        	  else do 
			                 c         <- getChar
				         case c of
				              '<' -> (do {(a,b,d,e,f) <- eatTag docs tags ct env w; ps a b d e f})
			   	              _   -> if not (getTT (head ct)) then ps (eatChar c (removeEnv env ct docs)) tags ct (setHeadNothing env) (addCharWs c w)
					                                      else (do
								                      (e,aw,eof') <- eatText [c] (updateAllWs c (head w))
									              if eof' then return (insert e (removeEnv env ct docs))
									                      else (do {
										                        (d',t',c',e',w') <- eatTag (insert e (removeEnv env ct docs)) tags ct (setHeadNothing env) (aw:tail w);
										      		         ps d' t' c' e' w'}))
removeEnv :: [Env] -> [Context] -> [Document] -> [Document]
removeEnv (Nothing:_) _ docs = docs
removeEnv ((Just dcd):_) (ct:_) (d:ds) = (enter dcd ct d) : ds

insert :: El -> [Document] -> [Document]
insert e (h:b) = (e : h) : b

setHeadNothing :: [Maybe a] -> [Maybe a]
setHeadNothing l@(Nothing:_) = l
setHeadNothing (_:l) = Nothing:l

addCharWs :: Char -> [Bool] -> [Bool]
addCharWs _ 	w@(False:_) 	= w
addCharWs ' '	w		= w
addCharWs '\t'	w		= w
addCharWs '\n'	w		= w
addCharWs '\r'	w		= w
addCharWs _	(_:w)		= False : w

updateAllWs :: Char -> Bool -> Bool
updateAllWs _		False	= False
updateAllWs ' '		x	= x
updateAllWs '\t'	x	= x
updateAllWs '\n'	x	= x
updateAllWs '\r'	x	= x
updateAllWs _		_	= False

eatText :: String -> Bool -> IO (El,Bool,Bool)
eatText l w = 
	do 
	   eof <- isEOF
	   if eof then return (Txt l,w, True)
	          else do
		         c <- getChar
			 case c of 
			      '<' -> return (Txt l, w,False)
			      _   -> eatText (c:l) (updateAllWs c w)



eatChar :: Char -> [Document] -> [Document] 

eatChar ' ' docs@((Ws:_):_) = docs -- compact whitespace
eatChar ' '  (h:t) = (Ws : h) : t 
eatChar '\n' docs@((Ws:_):_) = docs -- compact whitespace
eatChar '\n' (h:t) = (Ws : h) : t
eatChar '\t' docs@((Ws:_):_) = docs -- compact whitespace
eatChar '\t' (h:t) = (Ws : h) : t
eatChar '\r' docs@((Ws:_):_) = docs -- compact whitespace
eatChar '\r' (h:t) = (Ws : h) : t

eatChar c ( ( (Txt l) : hr) : tr) =  ((Txt (c:l)) : hr) : tr
eatChar c (h:t) = ((Txt [c]) : h) : t

eatTag :: [Document] -> [Tag] -> [Context] -> [Env] -> [Bool] -> IO ([Document],[Tag],[Context],[Env],[Bool])
eatTag docs tags ct env w =
  do 
     c <- getChar
     restTag
     return $ case c of
       		'/' -> popTag docs tags ct env w
       		'E' -> pushEM docs tags ct (Nothing:env)    (True:w)
       		'T' -> pushTT docs tags ct (Nothing:env)    (True:w)
       		'P' -> pushPlain docs tags ct (Nothing:env) (True:w)
       		_   -> pushTag c docs tags ct (Nothing:env) (True:w)
       
popTag :: [Document] -> [Tag] -> [Context] -> [Env] -> [Bool] -> ([Document],[Tag],[Context],[Env],[Bool])

--  CASE 1: no elements inside tag
popTag ([]:ds) (_:ts) (_:cs) (Nothing:es) (_:ws)= -- do not generate teg for empty document 
	(ds,ts,cs,es,ws)

-- CASE 2: one outstanding neighbour
--	SUBCASE A: no outstanding parent neighbour

popTag ([]:ds) (t:ts) (c1:c2:cs) (Just (dc,d):Nothing:es) (_:w:ws) =
	if c1 == c2 then (ds, ts, c2:cs, Just (dc,d):es,False:ws)
	            else (ds, ts, c2:cs, Just (enterDCTag t dc,d):es,False:ws)

--	SUBCASE B  an outstanding parent neighbour
popTag ([]:d:ds) (t:ts) (c1:c2:cs) (Just (dc1,d1) : Just (dc2,d2) : es) (_:w:ws) = 
	let dc1' 	= if c1 == c2 then dc1 else enterDCTag t dc1
	    result	= mergeByDeltaContext (dc2,d2) (dc1',d1) c2
	in case result of
		Left  (p2, p1) -> ((enter p2 c2 d): ds,ts,c2:cs,(Just p1):es,False:ws)
		Right p3       -> (d:ds,ts,c2:cs,(Just p3):es,False:ws)

-- CASE 3 elements inside tag without oustanding parent neighbour
--	SUBCASE A	no outsanding neighbour
--		SUBSUBCASE A	not all spaces
popTag (d1:d2:ds) (t:ts) (c1:c2:cs) (Nothing:Nothing:es) (False:w:ws) =
	if c1 == c2 then -- the current tqg does not add anything to the context
			 ((d1 ++ d2):ds, ts, c2:cs, Nothing:es,False:ws)
	            else -- the current tqg does add something to the context
		         (d2:ds, ts, c2:cs,Just (enterDCTag t emptyDeltaContext,d1):es,False:ws)

--		SUBSUBCASE B	all spaces
popTag (d1:d2:ds) (t:ts) (c1:c2:cs) (Nothing:Nothing:es) (_:w:ws) =
	if stdtfws t c2 then ((d1 ++ d2):ds,ts,c2:cs,Nothing:es,w:ws)
	                else if c1 == c2 then ((d1 ++ d2):ds, ts, c2:cs, Nothing:es,False:ws)
	                                 else (d2:ds, ts, c2:cs,Just (enterDCTag t emptyDeltaContext,d1):es,False:ws)

--	SUBCASE B	an outstanding neighbour
popTag (d1:d2:ds) (t:ts) (c1:c2:cs) (Just (dc,d) : Nothing : es) (_:w:ws) =
	if c1 == c2 then -- the current tag does not add anything to the context
			 ((d1 ++ d2):ds, ts, c2:cs, Just (dc,d) : es, False:ws)
	 	    else -- the current tag does add something to the context
		         if (enterDCTag t dc) == dc then ((Tag t d1:d2):ds,ts,c2:cs,Just (dc,d):es,False:ws)
		                                    else (d2:ds, ts, c2:cs, Just (enterDCTag t emptyDeltaContext, enter (dc,d) c1 d1) : es,False:ws)

-- CASE 4 elemens inside tag,, an outstanding parent neighbour
--	SUBCASE A	no outstanding neighbour
--		SUBSUBCASE A	not all spaces
popTag (d1:d2:ds) (t:ts) (c1:c2:cs) (Nothing:Just (dc,d):es) (False:w:ws) =
	if c1 == c2 then
			 ((d1 ++ (enter (dc,d) c2 d2)):ds,ts,c2:cs, Nothing:es,ws)
		    else
		    	 let result = mergeByDeltaContext (dc,d) (enterDCTag t emptyDeltaContext, d1) c2
			 in case result of
			         Left  (p2, p1) -> ((enter p2 c2 d2):ds,ts,c2:cs,Just p1:es,False:ws)
				 Right p3       -> (d2:ds,ts,c2:cs,Just p3 : es,False:ws)

--		SUBSUBCASE B	all spaces
popTag (d1:d2:ds) (t:ts) (c1:c2:cs) (Nothing:Just (dc,d):es) (_:w:ws) =
	if (stdtfws t c2) || (c1 == c2) then let result = mergeByDeltaContext (dc,d) (wssclone dc, d1) c2
					     in case result of
					     		Left  (p2, p1) -> ((d1 ++ (enter p2 c2 d2)):ds,ts,c2:cs,Nothing:es,False:ws)
							Right p3       -> (d2:ds,ts,c2:cs,Just p3 : es, False:ws)
		    			else let result = mergeByDeltaContext (dc,d) (enterDCTag t emptyDeltaContext, d1) c2
			 		     in case result of
			         			Left  (p2, p1) -> ((enter p2 c2 d2):ds,ts,c2:cs,Just p1:es,False:ws)
						 	Right p3       -> (d2:ds,ts,c2:cs,Just p3 : es,False:ws)

--	SUBCASE B	an outstanding neighbour
popTag (d1:d2:ds) (t:ts) (c1:c2:cs) (Just (dc1,doc1) : Just (dc2,doc2) : es) (_:ws) =
	if c1 == c2 then
			 ((d1 ++ (enter (dc2,doc2) c2 d2)):ds , ts, c2:cs,Just (dc1, doc1) : es, ws)
	            else if (enterDCTag t dc1) == dc1 then let result = mergeByDeltaContext (dc2,doc2) (enterDCTag t emptyDeltaContext, d1) c2
		    					   in case result of
							   		Left  (p2,p1) -> ((enter p1 c2 (enter p2 c2 d2)):ds,ts,c2:cs,Just (dc1,doc1):es,ws)
									Right p3      -> ((enter p3 c2 d2):ds,ts,c2:cs,Just (dc1,doc1):es,ws)
		         			      else let result = mergeByDeltaContext (dc2,doc2) (enterDCTag t emptyDeltaContext, enter (dc1, doc1) c1 d1) c2
			 			      in case result of
			 	 				Left  (p2, p1) -> ((enter p2 c2 d2) : ds, ts, c2:cs, Just p1: es, ws)
								Right p3       -> (d2 : ds, ts, c2:cs, Just p3 : es, ws)

{-
Nog te doen:
  De huidige neighbour bijhouden en daarmee proberen te mergen.
  Geen nieuwe DeltaContext -> Geen neighbour -> geen nieuwe neighbour, gewoon enteren
                           -> wel            -> oude neighbour enteren, gewoon enteren, geen nieuwe neighbour
  Wel nieuwe deltacontext (evt subneighbour enteren)
  	-> geen neighbour -> zelf wordt nieuwe neighbour
	-> wel neighbour -> proberen te enteren
				-> niet gelukt, neighbour enteren en zelf neighbour worden
				-> wel gelukt, merge wordt nieuwe neighbour
-}

enter :: (DeltaContext, Document) -> Context -> Document -> Document
enter (dc, doc) ct parent = let child = buildDCEl dc ct doc
                            in child ++ parent

pushEM :: [Document] -> [Tag] -> [Context] -> [Env] -> [Bool] -> ([Document],[Tag],[Context],[Env],[Bool])
pushEM docs tags (c:cs) env w = 
	([]:docs,EM:tags,(enterTag c EM):c:cs,env,w)
 
pushTT :: [Document] -> [Tag] -> [Context] -> [Env] -> [Bool] -> ([Document],[Tag],[Context],[Env],[Bool])
pushTT docs tags (c:cs) env w = 
	([]:docs,TT:tags,(enterTag c TT):c:cs,env,w)

pushPlain :: [Document] -> [Tag] -> [Context] -> [Env] -> [Bool] -> ([Document],[Tag],[Context],[Env],[Bool])
pushPlain docs tags (c:cs) env w = 
	([]:docs,PL:tags,(enterTag c PL):c:cs,env,w)
 
pushTag :: Char -> [Document] -> [Tag] -> [Context] -> [Env] -> [Bool] -> ([Document],[Tag],[Context],[Env],[Bool])
pushTag x docs tags (c:cs) env w = 
	let t = getTag x
	in ([]:docs,t:tags,(enterTag c t):c:cs,env,w)

restTag :: IO ()
restTag = 
  do 
     c <- getChar
     case c of 
       '>' -> return ()
       _   -> restTag

stdtfws :: Tag -> Context -> Bool
stdtfws TT _	= False
stdtfws U  _	= False
stdtfws _  ct	= (getU ct) == 0 -- catches the <U><PL> case
