module TextBasedOptimizer where

import Syntax
import Context
import List (delete)

tbo :: Document -> Document
tbo d = 
	let (dc,s) = textContextList d
	in  buildTree dc [[]] [([],[])]

textContextList :: Document -> ([(Document, Context)], Int)
textContextList doc = foldl (untree emptyContext) ([],0) doc

untree :: Context -> ([(Document, Context)],Int) -> El -> ([(Document, Context)],Int)
untree ct ([],c) Ws 		= ([([Ws],ct)],c)
untree ct ((h@(l,ct'):r),c) Ws	| ct == ct'	= ((Ws:l,ct'):r,c)
				| otherwise	= (([Ws],ct):h:r,c)
				
untree ct ([],c) t@(Txt _)			= ([([t],ct)],c)
untree ct ((h@(l,ct'):r),c) t@(Txt _)	| ct == ct' 	= ((t:l,ct'):r,c)
					| otherwise	= (([t],ct):h:r,c)

untree ct (p,c) (Tag t doc)	=
	let ct' 	= enterTag ct t
	in foldl (untree ct') (p,c+1) doc

untree _ _ _= error "untree"

--

buildTree ::
	[(Document, Context)]   -> -- next
	[[Tag]]			-> -- Tag increment stack
	[([Tag],Document)]	-> -- Tag increment + document
	Document

buildTree [] _ incstack = closeStack incstack

buildTree ((d,c):r) ts incstack =
	let t'		  	= genTags c
	in buildTree' d t' r ts incstack

buildTree _ _ _ = error "buildTree"
		
buildTree' d t' r (t:ts) incstack =
	let (different, new)	= sublist t t'
	in case (different,new) of
--		([],[]) -> buildTree 
		([],_)  -> buildTree  r (t':t:ts) ((new,d):incstack)
		(_,_)   -> buildTree' d t' r ts (pop incstack)

closeStack :: [([Tag],Document)] -> Document
closeStack [(_,d)] 	= d
closeStack s 		= closeStack $ pop s

pop :: [([Tag], Document)] -> [([Tag], Document)] 
pop ((t1,d1):(t2,d2):r) = 
	let e = buildElement t1 d1
	in ((t2,d2 ++ e):r)

pop [x] = error ("pop " ++ show x)

buildElement :: [Tag] -> Document -> Document
buildElement ts d = foldr (\x y -> [Tag x y]) d ts
	
sublist :: Eq a => [a] -> [a] -> ([a],[a])
sublist l  []		=	(l,[])
sublist [] l		= 	([],l)
sublist (x:xs) ys	| elem x ys	= sublist xs (delete x ys)
			| otherwise	= (x:xs, [])
